<?php
/**
 *  This helper class was made for the Copernica REST API and will
 *  handle GET, POST, PUT and DELETE calls. It can be used for all
 *  versions of the REST API.
 */

/**
 *  Helper class that represents the JWT
 */
require_once __DIR__.'/JSONWebToken.php';

/**
 *  Client class implementation
 */
class CopernicaRestAPI
{
    /**
     *  The access token
     *  @var    string
     */
    private $token;

    /**
     *  The version of the REST API
     *  @var    int
     */
    private $version;

    /**
     *  The API host
     *  @var    string
     */
    private $host = 'api.copernica.com';

    /**
     *  Cached JWT
     *  @var    JSONWebToken
     */
    private $jwt = null;

    /**
     *  Constructor
     *  @param  string      Access token
     *  @param  int         Version parameter (optional)
     */
    public function __construct($token, $version = 4)
    {
        // copy the token
        $this->token = $token;

        // set the version
        $this->version = $version;
    }

    /**
     *  Do a GET request
     *
     *  @param  string      Resource to fetch
     *  @param  array       Associative array with additional parameters
     *  @return array       Associative array with the result
     */
    public function get($resource, array $parameters = [])
    {
        // versions 4 and newer use JWT authorization
        if ($this->version >= 4)
        {
            // get the JWT for the auth header
            if (is_null($jwt = $this->jwt())) return null;

            // the query string
            $query = http_build_query($parameters);

            // construct curl resource
            $curl = curl_init("https://{$this->host}/v{$this->version}/$resource?$query");

            // additional options
            curl_setopt_array($curl, [CURLOPT_HTTPHEADER => ["Authorization: Bearer {$jwt}"]]);
        }

        // older versions include the access_token in the url
        else
        {
            // the query string
            $query = http_build_query(['access_token' => $this->token] + $parameters);

            // construct curl resource
            $curl = curl_init("https://{$this->host}/v{$this->version}/$resource?$query");
        }

        // additional options
        curl_setopt_array($curl, [CURLOPT_RETURNTRANSFER => true]);

        // do the call
        $answer = curl_exec($curl);

        // do we have a JSON output? we can be nice and parse it for the user
        if (curl_getinfo($curl, CURLINFO_CONTENT_TYPE) == 'application/json') {

            // the JSON parsed output
            $jsonOut = json_decode($answer, true);

            // if we have a json error then we have some garbage in the out
            if (json_last_error() != JSON_ERROR_NONE) return null;

            // return the json
            return $jsonOut;
        }

        // clean up curl resource
        curl_close($curl);

        // it's not JSON so we out it just like that
        return $answer;
    }

    /**
     *  Execute a POST request.
     *
     *  @param  string      Resource name
     *  @param  array       Associative array with data to post
     *
     *  @return mixed       ID of created entity, or simply true/false
     *                      to indicate success or failure
     */
    public function post($resource, array $data = [])
    {
        // Pass the request on
        return $this->sendData($resource, $data, [], "POST");
    }

    /**
     *  Execute a PUT request.
     *
     *  @param  string      Resource name
     *  @param  array       Associative array with data to post
     *  @param  array       Associative array with additional parameters
     *
     *  @return mixed       ID of created entity, or simply true/false
     *                      to indicate success or failure
     */
    public function put($resource, $data, array $parameters = [])
    {
        // Pass the request on
        return $this->sendData($resource, $data, $parameters, "PUT");
    }

    /**
     *  Execute a DELETE request
     *
     *  @param  string      Resource name
     *  @param  array       The data for the request body
     *  @param  array       The parameters for in the URL
     *  @return bool
     */
    public function delete($resource, $data = [], $parameters = [])
    {
        // Pass the request on
        return $this->sendData($resource, $data, $parameters, "DELETE");
    }

    /**
     *  Execute a request to create/edit data. (PUT + POST)
     *
     *  @param  string      Resource name
     *  @param  array       Associative array with data to post
     *  @param  array       Associative array with additional parameters
     *  @param  string      Method to use (POST or PUT)
     *
     *  @return mixed       ID of created entity, or simply true/false
     *                      to indicate success or failure
     */
    public function sendData($resource, array $data = [], array $parameters = [], $method = "POST")
    {
        // data will be json encoded
        $data = json_encode($data);

        // define headers to always include
        $headers = ['content-type: application/json', 'content-length: '.strlen($data)];

        // versions 4 and newer use JWT authorization
        if ($this->version >= 4)
        {
            // get the JWT for the auth header
            if (is_null($jwt = $this->jwt())) throw new Exception('JWT unavailable. Try again later.');

            // the query string
            $query = http_build_query($parameters);

            // append the authorization header
            $headers[] = 'Authorization: Bearer '.$jwt;
        }

        // older versions include the access_token in the url
        else $query = http_build_query(['access_token' => $this->token] + $parameters);

        // construct curl resource
        $curl = curl_init("{$this->host}/v{$this->version}/$resource?$query");

        // set the default options
        $options = [
            CURLOPT_HEADER          =>  true,
            CURLOPT_RETURNTRANSFER  =>  true,
            CURLOPT_HTTPHEADER      =>  $headers,
            CURLOPT_POSTFIELDS      =>  $data
        ];

        // for POST we have a special option
        if ($method == "POST") $options[CURLOPT_POST] = true;

        // otherwise use customrequest option
        else $options[CURLOPT_CUSTOMREQUEST] = $method;

        // additional options
        curl_setopt_array($curl, $options);

        // execute the call
        $answer = curl_exec($curl);

        // retrieve the HTTP status code
        $httpCode = curl_getinfo($curl, CURLINFO_HTTP_CODE);

        // clean up curl resource
        curl_close($curl);

        // bad request
        if (!$httpCode || $httpCode == 400) return false;

        // try and get the X-Created id from the header
        // if we have none we just return true for a succesful request
        if (!preg_match('/X-Created:\s?(\d+)/i', $answer, $matches)) return true;

        // return the id of the created item
        return $matches[1];
    }

    /**
     *  Helper to fetch the JWT used in versions 4 and up of the API.
     *  @return string|null
     */
    public function jwt(): ?string
    {
        // load from class member
        $jwt = $this->jwt;

        // if the JWT wasn't stored yet or has already expired, request a new one
        if (!is_object($jwt) || $jwt->expired()) $jwt = JSONWebToken::fromServer($this->token);

        // the request to retrieve the token migh have failed
        if (!is_object($jwt)) return $this->jwt = null;

        // cache in a class member
        $this->jwt = $jwt;

        // return raw contents (the JWT string)
        return $jwt->raw();
    }
}
